# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a Fairyfly Model, which can be used for simulation.
-

    Args:
        rooms_: A list of fairyfly Shapes to be added to the Model. Note that at
            least one Shape is necessary to make a simulate-able THERM model.
        boundaries_: A list of fairyfly Boundary objects to be added to the Model.
            Note that at least two distinct boundary conditions are necessary
            to make a simulate-able THERM model.
        _name_: Text to be used for the Model name. If no name is provided, it
            will be "unnamed".

    Returns:
        report: Reports, errors, warnings, etc.
        model: A Fairyfly Model object possessing all of the input geometry
            objects.
"""

ghenv.Component.Name = 'FF Model'
ghenv.Component.NickName = 'Model'
ghenv.Component.Message = '1.9.1'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '0 :: Create'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:  # import the core fairyfly dependencies
    from fairyfly.model import Model
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
    from ladybug_rhino.config import units_system, current_tolerance, angle_tolerance
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def check_all_geo_none():
    """Check whether all of the geometry inputs to this component are None."""
    return all(obj_list == [] or obj_list == [None] for obj_list in
               (shapes_, boundaries_))


if all_required_inputs(ghenv.Component) and not check_all_geo_none():
    # check the Rhino Model units
    units, tolerance = units_system(), current_tolerance()
    tol_msg = Model.check_reasonable_tolerance(units, tolerance)
    if tol_msg is not None:
        give_warning(ghenv.Component, tol_msg)

    # create the model
    model = Model(
        shapes_, boundaries_,
        units=units, tolerance=tolerance, angle_tolerance=angle_tolerance)
    model.display_name = _name_ if _name_ is not None else 'unnamed'



# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Search for available Materials within the fairyfly therm standards library.
-

    Args:
        keywords_: Optional keywords to be used to narrow down the output list of
            materials. If nothing is input here, all available materials
            will be output.
        join_words_: If False or None, this component will automatically split
            any strings of multiple keywords (spearated by spaces) into separate
            keywords for searching. This results in a greater liklihood of
            finding an item in the search but it may not be appropropriate for
            all cases. You may want to set it to True when you are searching for
            a specific phrase that includes spaces. Default: False.

    Returns:
        solid_mats: A list of solid materials within the fairyfly therm standards
            library (filtered by keywords_ if they are input).
        cavity_mats: A list of gas cavity materials within the fairyfly therm standards
            library (filtered by keywords_ if they are input).
"""

ghenv.Component.Name = 'FF Search Materials'
ghenv.Component.NickName = 'SearchMats'
ghenv.Component.Message = '1.9.0'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '1 :: THERM'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:  # import the fairyfly-core dependencies
    from fairyfly.search import filter_array_by_keywords
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly:\n\t{}'.format(e))

try:  # import the fairyfly-therm dependencies
    from fairyfly_therm.lib.materials import SOLID_MATERIALS
    from fairyfly_therm.lib.materials import CAVITY_MATERIALS
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly_therm:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import turn_off_old_tag
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))
turn_off_old_tag(ghenv.Component)


if len(keywords_) == 0:
    solid_mats = sorted(SOLID_MATERIALS)
    cavity_mats = sorted(CAVITY_MATERIALS)
else:
    split_words = True if join_words_ is None else not join_words_
    solid_mats = sorted(filter_array_by_keywords(SOLID_MATERIALS, keywords_, split_words))
    cavity_mats = sorted(filter_array_by_keywords(CAVITY_MATERIALS, keywords_, split_words))



# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a custon radiant environment for Fairyfly boundary.
_
Assigning values here will create radiant conditions that are different from
stanbdard NFRC conditions (where radiant temperature equals air temperature,
the emissivity of the environment is 1, and there is no additional heat flux).
-

    Args:
        _geo: Planar Polyline or Line geometry representing the boundary.
        rad_temp_: A number for the radiant temperature at the boundary
            in degrees Celsius. If None, this will be the same as the specified
            air temperature. (Default: None).
        emissivity_: An optional number between 0 and 1 to set the emissivity
            along the boundary, which represents the emissivity of the
            environment to which the material in contact with the boundary is
            radiating to. (Default: 1).
        heat_flux_: An optional number in W/m2 that represents additional energy
            flux across the boundary. This can be used to account for solar flux
            among other forms of heat flux. (Default: 0).

    Returns:
        report: Reports, errors, warnings, etc.
        rad_env: Radiant environment properties that can be plugged into the "FF Boundary"
            component in order to customize the radiant environment within
            THERM simulation.
"""

ghenv.Component.Name = 'FF Radiant Environment'
ghenv.Component.NickName = 'RadEnv'
ghenv.Component.Message = '1.9.0'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '0 :: Create'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs, objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    emiss = emissivity_ if emissivity_ is not None else 1.0
    heat_flux = heat_flux_ if heat_flux_ is not None else 0
    rad_env = objectify_output('Radiant Enviornment', [emiss, rad_temp_, heat_flux])



# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create Fairyfly Boundary.
-

    Args:
        _thmz: Path to a THMZ file that has been simulated in THERM. This can be the
            direct output of the "FF Model to THMZ" component as long as
            run_ has been set to True.
        _u_type_: Optional text or an integer to set the type of U-Factor to import from
            the THMZ file if something other than the U-Factor along the total
            boundary length is desired. Choose from the following options. (Default: Total)
                0 - Total
                1 - Projected X
                2 - Projected Y
                3 - Projected in Glass Plane
                4 - Custom Rotation
        ip_: Set to True to have all data imported with IP Units (Btu/h-ft2-F and in)
            instead of the default SI units of W/m2-K and mm.

    Returns:
        report: Reports, errors, warnings, etc.
        tags: The names of each of the U-Factor tags in the THMZ file.
        u_factors: The U-Factors accross each boundary that was labeled with a U-Factor tag.
            This will be in W/m2-K by default unless ip_ is set to True in which
            case it will be in Btu/h-ft2-F.
        lengths: The proejcted lengths of each boundary that was labeled with a U-Factor tag.
            This will be the total length of the boundary if the default u_type
            of "Total" has been used and will account for the length only in the
            projection plane if a different u_type is used. Values are in mm by
            default unless ip_ is set to True, in which case it will be in inches.
"""

ghenv.Component.Name = 'FF Read UFactor Result'
ghenv.Component.NickName = 'UFactorResult'
ghenv.Component.Message = '1.9.0'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '1 :: THERM'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:  # import the core fairyfly dependencies
    from fairyfly_therm.result import THMZResult
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly:\n\t{}'.format(e))


try:  # import the core ladybug dependencies
    from ladybug.datatype.uvalue import UValue
    from ladybug.datatype.distance import Distance
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

U_TYPES = {
    'total': 'total',
    '0': 'total',
    'projected x': 'projected_x',
    '1': 'projected_x',
    'projected y': 'projected_y',
    '2': 'projected_y',
    'projected in glass plane': 'projected_in_glass_plane',
    '3': 'projected_in_glass_plane',
    'custom rotation': 'custom_rotation',
    '4': 'custom_rotation'
}


if all_required_inputs(ghenv.Component):
    # create the THMZResult object and check to be sure that it has been simulated
    result_obj = THMZResult(_thmz)
    uf_objs = result_obj.u_factors
    if uf_objs is None:
        msg = 'No U-Factor data was found within the THMZ file.\nMake sure that the THMZ ' \
            'file has been successfully simulated in THERM.'
        print(msg)
        give_warning(ghenv.Component, msg)

    elif len(uf_objs) == 0:
        msg = 'U-Factor tags were not assigned to any boundaries.\nMake sure that ' \
            'a u_factor_tag_ has been input to one of the "FF Boundary" components.'
        print(msg)
        give_warning(ghenv.Component, msg)

    else:
        # figure out the right attributes to get from the UFactor objects
        if _u_type_ is None:
            b_attr = 'total'
        else:
            b_attr = U_TYPES[_u_type_.lower()]
        u_attr = '{}_u_factor'.format(b_attr)
        l_attr = '{}_length'.format(b_attr)
        
        # extract all of the U-Factor information
        tags, u_factors, lengths = [], [], []
        for uf_obj in uf_objs:
            tags.append(uf_obj.name)
            u_factors.append(getattr(uf_obj, u_attr))
            lengths.append(getattr(uf_obj, l_attr))

        # if IP units are requested, convert them
        if ip_:
            u_factors = UValue().to_ip(u_factors, 'W/m2-K')[0]
            lengths = Distance().to_ip(lengths, 'mm')[0]



# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create a custom solid material, which can be plugged into the "FF Shape" component
to change the properties of the material in a THERM simulation.
-

    Args:
        _name_: Text to set the name for the material.
        _conductivity: Number for the thermal conductivity of the material [W/m-K].
        _emissivity_: Number between 0 and 1 for the infrared hemispherical
            emissivity of the material. (Default: 0.9).
        density_: Number for the density of the material [kg/m3].
        porosity_: Optional number between 0 and 1 for the porosity of the material.
        spec_heat_: Number for the specific heat of the material [J/kg-K].
        vapor_resist_: Optional number for the water vapor diffusion resistance factor.
        rgb_color_: An optional color to set the color of the material when it is
            imported to THERM. If unspecified, a randomly-generated color is assigned.

    Returns:
        mat: A standard solid material that can be assigned to a Fairyfly Shape
            to change the properties of the material in a THERM simulation.
"""

ghenv.Component.Name = 'FF Solid Material'
ghenv.Component.NickName = 'SolidMat'
ghenv.Component.Message = '1.9.1'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '1 :: THERM'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import uuid

try:  # import the core ladybug dependencies
    from ladybug.color import Color
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:  # import the fairyfly-therm dependencies
    from fairyfly_therm.material.solid import SolidMaterial
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly_therm:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the default material properties
    _emissivity_ = 0.9 if _emissivity_ is None else _emissivity_

    # create the material
    mat = SolidMaterial(
        _conductivity, _emissivity_, None,
        density_, porosity_, spec_heat_, vapor_resist_
    )
    mat.display_name = _name_ if _name_ is not None else \
        'Solid Material {}'.format(str(uuid.uuid4())[:8])
    if rgb_color_ is not None:
        mat.color = Color(rgb_color_.R, rgb_color_.G, rgb_color_.B)



# Fairyfly: A Plugin for Environmental Analysis (GPL)
# This file is part of Fairyfly.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Fairyfly; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Color Fairyfly Shapes in the Rhino scene using their attributes.
_
This can be used as a means to check that correct properties are assigned to
different shapes.
-

    Args:
        _ff_objs: An array of fairyfly Shapes to be colored with their attributes
            in the Rhino scene. This can also be an entire Model to be colored.
        _attribute_: Text for the name of the attribute with which the shapes should
            be colored. The "FF Shape Attributes" component lists all of the
            attributes of the shape. (Default: THERM Material).
        legend_par_: An optional LegendParameter object to change the display
            of the colored shapes. (Default: None).

    Returns:
        wire_frame: A list of lines representing the outlines of the _ff_objs.
        legend: Geometry representing the legend for colored meshes.
        mesh: Meshes of the shapes colored according to their attributes.
        values: A list of values noting the attribute assigned to each shape.
        colors: A list of colors noting the color of each shape in the Rhino scene.
            This can be used in conjunction with the native Grasshopper
            "Custom Preview" component to create custom visualizations in
            the Rhino scene.
        vis_set: An object containing VisualizationSet arguments for drawing a detailed
            version of the ColorRoom in the Rhino scene. This can be connected to
            the "LB Preview Visualization Set" component to display this version
            of the visualization in Rhino.
"""

ghenv.Component.Name = 'FF Color Shape Attributes'
ghenv.Component.NickName = 'ColorShapeAttr'
ghenv.Component.Message = '1.9.2'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '0 :: Create'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:  # import the core fairyfly dependencies
    from fairyfly.model import Model
    from fairyfly.shape import Shape
    from fairyfly.colorobj import ColorShape
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.fromgeometry import from_face3ds_to_colored_mesh, \
        from_face3d_to_wireframe
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.color import color_to_color
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

DEFAULT_PROP = 'properties.therm.material.display_name'


if all_required_inputs(ghenv.Component):
    # extract any shapes from input Models
    _attribute_ = DEFAULT_PROP if _attribute_ is None else _attribute_
    shapes = []
    for ff_obj in _ff_objs:
        if isinstance(ff_obj, Model):
            shapes.extend(ff_obj.shapes)
        elif isinstance(ff_obj, Shape):
            shapes.append(ff_obj)

    # create the ColorShape visualization object
    color_obj = ColorShape(shapes, _attribute_, legend_par_)
    # assign material colors if this is what is being used for coloring
    if _attribute_ == DEFAULT_PROP and color_obj.legend_parameters.are_colors_default:
        color_model = Model(shapes)
        color_dict = {mat.display_name: mat.color
                      for mat in color_model.properties.therm.materials}
        leg_colors = [color_dict[m_name] for m_name in color_obj.attributes_unique]
        if len(leg_colors) == 1:
            leg_colors = [leg_colors[0]] * 2
        color_obj.legend_parameters.colors = leg_colors

    # output the visualization geometry
    graphic = color_obj.graphic_container
    values = color_obj.attributes_original
    flat_geo = color_obj.geometry
    mesh = []
    for fc, col in zip(flat_geo, graphic.value_colors):
        mesh.append(from_face3ds_to_colored_mesh([fc], col))
    wire_frame = []
    for face in flat_geo:
        wire_frame.extend(from_face3d_to_wireframe(face))
    legend = legend_objects(graphic.legend)
    colors = [color_to_color(col) for col in graphic.value_colors]
    vis_set = color_obj



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2025, Ladybug Tools.
# You should have received a copy of the GNU Affero General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license AGPL-3.0-or-later <https://spdx.org/licenses/AGPL-3.0-or-later>

"""
Create parameters with criteria for meshing THERM geometry.
-

    Args:
        _mesh_type_: Text to indicate the type of meshing algorithm to use. Choose from
            the following. Simmetrix is generally more flexible and capable of
            handling more complex geometry when compared with the QuatTree. However,
            the structure of QuadTree meshes is more predictable. (Default: Simmetrix).
            * Simmetrix
            * QuadTree
        _parameter_: A positive integer for the minimum number of subdivisions to
            be performed while meshing the input geometry. The higher the mesh
            control parameter, the smaller the maximum size of finite elements
            in the model and the smoother the results will appear. However, higher
            mesh parameters will also require more time to run. (Default: 20).
        _check_error_: Boolean to note whether the error estimator should
            be run as part of the finite element analysis. If the global error
            is above a specified value, then the error estimator signals the mes
            generator, and the mesh is refined in areas where the potential
            for error is high.  The refined mesh is sent back to the finite
            element solver, and a new solution is obtained. (Default: True).
        _max_error_: A number between 0 and 100 for the percent error energy
            norm used by the error estimator. This is the maximum value
            of the error energy divided by the energy of the sum of the
            recovered fluxes and the error, multiplied by 100. (Default: 10).
        _max_iter_: A positive integer for the number of iterations between the error
            estimator and the solver to be performed before the finding
            a solution is abandoned and the program exits. (Default: 5).

    Returns:
        meshing: Parameters with criteria for creating the finite element mesh.
            These can be connected to the "FF Model to THMZ" component in
            order to specify settings for the THERM simulation.
"""

ghenv.Component.Name = 'FF Meshing Control'
ghenv.Component.NickName = 'Meshing'
ghenv.Component.Message = '1.9.0'
ghenv.Component.Category = 'Fairyfly'
ghenv.Component.SubCategory = '1 :: THERM'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from fairyfly_therm.simulation.mesh import MeshControl
except ImportError as e:
    raise ImportError('\nFailed to import fairyfly_therm:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import turn_off_old_tag
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))
turn_off_old_tag(ghenv.Component)

# set default sizing factors
_mesh_type_ = 'Simmetrix' if _mesh_type_ is None else _mesh_type_
_parameter_ = 20 if _parameter_ is None else _parameter_
_check_error_ = True if _check_error_ is None else _check_error_
_max_error_ = 10 if _max_error_ is None else _max_error_
_max_iter_ = 5 if _max_iter_ is None else _max_iter_

# create the object
meshing = MeshControl(_mesh_type_, _parameter_, _check_error_, _max_error_, _max_iter_)
